#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>



/* defines */
#include "auc_defs.h"
/* reading data files */
#include "data_access.h"
/* vector utilities */
#include "vect_utils.h"
/* weight scaling */
#include "scaling.h"
/* attribute normalization */
#include "attribute_norms.h"
/* exact AUC computation */
#include "exact_auc.h"
/* options parsing */
#include "opt_parse.h"
/* minimization along the gradient */
#include "minim_along_grad.h"
/* chebyshev expansion */
#include "cheb_expand.h"


/* options */
struct options opt;


/* sigmoid function for soft-AUC */
static FLT_TYPE beta_tmp; /* current value of beta.  As scale of
			     weights can change, we are no longer
			     invariant to scaling, and beta needs to
			     adapt to take care of this effect. */

static FLT_TYPE
sigmoid(FLT_TYPE x)
{
    return (FLT_TYPE)1 / ((FLT_TYPE)1 + EXP_F(-beta_tmp * x));
}


/* 
 * initialize Chebyshev expansion coefficients
 */
static void
init_cheb_coeff(FLT_TYPE * step_coeff, size_t maxdeg)
{
    size_t i, j;
    FLT_TYPE * cheby_exp_coeff;	/* coefficients of Chebyshev expansion */
    FLT_TYPE * cheby_coeff1, * cheby_coeff2; /* coefficients of Chebyshev polynomials */

    /* for soft-AUC */
    if(opt.soft_auc)
    {
	cheby_exp_coeff = cheb_expand(sigmoid, maxdeg);
    }
    else
    {
	/* Chebyshev expansion of step function */
	cheby_exp_coeff = malloc(sizeof(FLT_TYPE) * (maxdeg + 1));
	cheby_exp_coeff[0] = (FLT_TYPE)0.5;
	cheby_exp_coeff[1] = _2_PI;
	for(i = 2; i <= maxdeg; i++)
	{
	    FLT_TYPE c;
	    /* i-th Chebyshev coefficient of the step function */
	    if(i % 2 == 0)
	    {
		c = (FLT_TYPE)0;
	    }
	    else
	    {
		c = (FLT_TYPE)1 / i;
		if(i % 4 == 3)
		{
		    c = -c;
		}
		c /= PI_2;
	    }
	    cheby_exp_coeff[i] = c;
	}
    }

    /* convert Chebyshev expansion into a standard polynomial. */
    cheby_coeff1 = malloc(sizeof(FLT_TYPE) * (maxdeg + 1));
    cheby_coeff2 = malloc(sizeof(FLT_TYPE) * (maxdeg + 1));
    for(i = 0; i <= maxdeg; i++)
    {
	step_coeff[i] = (FLT_TYPE)0;
	cheby_coeff1[i] = (FLT_TYPE)0;
	cheby_coeff2[i] = (FLT_TYPE)0;
    }
    cheby_coeff1[1] = (FLT_TYPE)1;
    cheby_coeff2[0] = (FLT_TYPE)1;
    step_coeff[0] = cheby_exp_coeff[0];
    step_coeff[1] = cheby_exp_coeff[1];
    for(i = 2; i <= maxdeg; i++)
    {
	FLT_TYPE *swap_tmp;

	/* compute coefficients of the next Chebyshev polynomial */
	cheby_coeff2[0] = -cheby_coeff2[0];
	for(j = 1; j <= i; j++)
	{
	    cheby_coeff2[j] = 2 * cheby_coeff1[j-1] - cheby_coeff2[j];
	}
	swap_tmp = cheby_coeff2;
	cheby_coeff2 = cheby_coeff1;
	cheby_coeff1 = swap_tmp;
	/* update the polynomial */
	for(j = 0; j <= i; j++)
	{
	    step_coeff[j] += cheby_coeff1[j] * cheby_exp_coeff[i];
	}
    }
    free(cheby_exp_coeff);
    free(cheby_coeff2);
    free(cheby_coeff1);
}

/*
 * allocate and initialize arrays
 */
void
init_arrays(size_t nattrs, size_t maxdeg, FLT_TYPE ** step_coeff,
	    FLT_TYPE ** weights, FLT_TYPE ** grad)
{
    size_t i;
    /* coefficients of step function approximation */
    *step_coeff = malloc(sizeof(FLT_TYPE) * (maxdeg + 1));


    init_cheb_coeff(*step_coeff, maxdeg);


    /* initialize weights */
    *weights = malloc(sizeof(FLT_TYPE) * nattrs);
    //drand48();
    //drand48();
    //drand48();
    for(i = 0; i < nattrs; i++)
    {
	(*weights)[i] = (FLT_TYPE)1;
	//(*weights)[i] = (FLT_TYPE)drand48();
    }
    /* gradient */
    *grad = malloc(sizeof(FLT_TYPE) * nattrs);
}
void
free_arrays(FLT_TYPE * step_coeff, FLT_TYPE * weights, FLT_TYPE * grad)
{
    free(grad);
    free(weights);
    free(step_coeff);
}





FLT_TYPE
compute_gradient(const FLT_TYPE * weights, FLT_TYPE * grad,
		 struct minim_cache * mc,
		 size_t nattrs, size_t maxdeg,
		 FILE * f)
{
    FLT_TYPE * X;
    FLT_TYPE * supports_f0;	/* supports of powers of f */
    FLT_TYPE * supports_fx0;	/* supports of powers of f times x_i */
    FLT_TYPE * supports_f1;	/* supports of powers of f */
    FLT_TYPE * supports_fx1;	/* supports of powers of f times x_i */
    FLT_TYPE * supports_f;	/* supports of powers of f */
    FLT_TYPE * supports_fx;	/* supports of powers of f times x_i */
    int class;
    size_t i, j;
    size_t ind;
    FLT_TYPE fpow;

    struct minmax mm;


    X = malloc(sizeof(FLT_TYPE) * nattrs);

    /* allocate memory for supports */
    supports_f0  = malloc(sizeof(FLT_TYPE) * (maxdeg + 1));
    supports_fx0 = malloc(sizeof(FLT_TYPE) * (maxdeg + 1) * nattrs);
    supports_f1  = malloc(sizeof(FLT_TYPE) * (maxdeg + 1));
    supports_fx1 = malloc(sizeof(FLT_TYPE) * (maxdeg + 1) * nattrs);
    for(i = 0; i <= maxdeg; i++)
    {
	supports_f0[i] = (FLT_TYPE)0;
	supports_f1[i] = (FLT_TYPE)0;
    }
    for(i = 0; i < (maxdeg + 1) * nattrs; i++)
    {
	supports_fx0[i] = (FLT_TYPE)0;
	supports_fx1[i] = (FLT_TYPE)0;
    }

    /* compute supports needed for gradient */
    rewind(f);
    init_minmax(&mm);
    while((class = read_and_normalize_record(f, X, nattrs)) != -1)
    {
	FLT_TYPE f;
	FLT_TYPE rescale_factor; /* how much to rescale supports
				    during current iteration to
				    account for increase in
				    max_M */
	FLT_TYPE rescale_pow;	 /* powers of the rescaling factor
				    used to scale current values of
				    supports */
	int do_rescale;

	///* print the record */
	//vect_print(X, nattrs);
	//printf("class: %d\n", class);

	if(class == 0)
	{
	    supports_f = supports_f0;
	    supports_fx = supports_fx0;
	}
	else
	{
	    supports_f = supports_f1;
	    supports_fx = supports_fx1;
	}

	f = compute_f(weights, X, nattrs);

	do_rescale = update_minmax(&mm, f, class);
	f *= mm.scale_factor;
	rescale_factor = mm.rescale_factor;

	/* rescale supports for both classes */
	if(do_rescale)
	{
	    rescale_pow = (FLT_TYPE)1;
	    for(j = 0; j <= maxdeg; j++)
	    {
		supports_f0[j] *= rescale_pow;
		supports_f1[j] *= rescale_pow;
		rescale_pow *= rescale_factor;
	    }
	    ind = 0;
	    for(i = 0; i < nattrs; i++)
	    {
		rescale_pow = (FLT_TYPE)1;
		
		for(j = 0; j <= maxdeg; j++)
		{
		    supports_fx0[ind] *= rescale_pow;
		    supports_fx1[ind] *= rescale_pow;
		    ind++;
		    rescale_pow *= rescale_factor;
		}
	    }
	}
	/* update supports for current class */
	ind = 0;
	for(i = 0; i < nattrs; i++)
	{
	    FLT_TYPE fpow_x = X[i];
	
	    for(j = 0; j <= maxdeg; j++)
	    {
		supports_fx[ind] += fpow_x;
		ind++;
		fpow_x *= f;
	    }
	}
	fpow = (FLT_TYPE)1;
	for(j = 0; j <= maxdeg; j++)
	{
	    supports_f[j] += fpow;
	    fpow *= f;
	}
    }

    //printf("gradient scale "PFMT"\n", (FLT_TYPE)1 / mm.scale_factor);

    /* for soft-AUC need to regenerate coefficients to reflect new
       scale. */
    if(opt.soft_auc)
    {
	beta_tmp = opt.beta / mm.scale_factor;
	init_cheb_coeff(mc -> step_coeff, maxdeg);
	precompute_alphas(mc, mc -> step_coeff, maxdeg);
    }

    /* approximate gradient */
    for(i = 0; i < nattrs; i++)
    {
	FLT_TYPE AUC_deriv = (FLT_TYPE)0;
	size_t k, l;
	for(k = 0; k <= maxdeg; k++)
	{
	    /* skip zero coefficients */
	    if(k > 0 && k % 2 == 0)
	    {
		continue;
	    }

	    for(l = 0; l <= k; l++)
	    {
		FLT_TYPE alpha_kl = mc -> alphas[k*(k+1)/2 + l];;
		FLT_TYPE sup0a, sup1a;
		FLT_TYPE sup0b, sup1b;

    	        if(l > 0)
		{
                    sup0a = supports_f0[k-l];
                    sup1a = supports_fx1[i * (maxdeg + 1) + l-1];
    	            AUC_deriv += alpha_kl * l * sup0a * sup1a;
		}
    	        if(l < k)
		{
		    sup0b = supports_fx0[i * (maxdeg + 1) + k-l-1];
		    sup1b = supports_f1[l];
		    AUC_deriv += alpha_kl * (k-l) * sup0b * sup1b;
		}
	    }
	}
	grad[i] = AUC_deriv;
    }

    free(supports_fx1);
    free(supports_f1);
    free(supports_fx0);
    free(supports_f0);
    free(X);

    return mm.scale_factor;
}




int main(int argc, char* argv[])
{
    FILE * f;
    size_t nattrs;
    size_t iter;

    size_t N0, N1;		/* number of cases in class 0 and 1 */

    FLT_TYPE * step_coeff;	/* polynomial coefficients of the				   step function */
    FLT_TYPE * weights;		/* the vector of weights */
    FLT_TYPE * grad;		/* the gradient */

    /* best values in given iteration */
    FLT_TYPE best_auc = (FLT_TYPE) -1;
    FLT_TYPE best_alpha;
    FLT_TYPE prev_auc;		/* previous value of auc (for stopping
				   criterion) */

    struct minim_cache mc;	/* cache for minimization along gradient */

    clock_t start_time, end_time; /* timing */


    start_time = clock();

    parse_options(&opt, argc, argv);

    f = fopen(opt.train_fname, "r");
    if(f == NULL)
    {
	printf("Cannot open file %s\n\n",opt.train_fname);
	return 1;
    }

    nattrs = get_nattrs(f);
    if(nattrs == 0UL)
    {
	printf("Problem reading file %s\n\n", opt.train_fname);
	return 1;
    }
    nattrs--; 			/* exclude class attribute */
    printf("%ld attributes (excluding class)\n", nattrs);

    /* cache file in memory */
    if(opt.cache_data)
    {
	cache_file(f, nattrs);
    }


    init_arrays(nattrs, opt.maxdeg, &step_coeff, &weights, &grad);
    init_minim_cache(&mc, opt.maxdeg, opt.lin_opt_step, step_coeff);

    /* normalize all x's */
    init_attribute_normalization(f, nattrs, &N0, &N1, opt.norm);

    /* optimizer iterations */
    for(iter = 0; iter < opt.maxiter; iter++)
    {
	size_t i;
	size_t lin_opt_step;

	/* support needed for minimization along the gradient */
	FLT_TYPE * supports_fwfg0;
	FLT_TYPE * supports_fwfg1;
	FLT_TYPE auc, alpha;
	FLT_TYPE scale;
	FLT_TYPE scale_weights, scale_grad;


	printf("Iteration %ld\n", iter + 1);
	if(opt.soft_auc || opt.regul)
	{
	    vect_normalize(weights, weights, nattrs);
	}
	scale = compute_gradient(weights, grad, &mc, nattrs, opt.maxdeg, f);
	/* printf("grad . W = "PFMT"\n",
	   vect_check_orthogonality(weights, grad, nattrs)); */


	/* add regularization term */
	if(opt.regul)
	{
	    FLT_TYPE weights_cur[nattrs];
	    //FLT_TYPE *weights_cur = weights;
	    FLT_TYPE regul = (FLT_TYPE)0;	/* regularization penalty */

	    vect_normalize(weights, weights_cur, nattrs);
	    for(i = 0; i < nattrs; i++)
	    {
		weights_cur[i] *= 1;
	    }

	    for(i = 0; i < nattrs; i++)
	    {
		if(opt.rumelhart)
		{
		    FLT_TYPE tmp = weights_cur[i] * weights_cur[i];
		    regul = (FLT_TYPE)2 * weights_cur[i] / (tmp * tmp + (FLT_TYPE)2 * tmp + (FLT_TYPE)1);
		}
		else if(opt.laplace)
		{
		    FLT_TYPE tmp = weights_cur[i] < 0 ? (FLT_TYPE)-1 : (FLT_TYPE)1 ;
		    regul = tmp;
		}
		grad[i] -= opt.lambda * regul * N0 * N1;
		//!grad[i] = -opt.lambda * regul;
	    }
	    //vect_print(weights_cur, nattrs);
	    //vect_print(grad, nattrs);
	    //vect_project_on_orthogonal_space(grad, weights, grad, nattrs);
	    //vect_print(grad, nattrs);
	    //printf("grad . W = "PFMT"\n",
	    //   vect_check_orthogonality(weights, grad, nattrs));
	}


	/* project gradient on space orthogonal to weights */
	if(opt.soft_auc || opt.regul)
	{
	    vect_project_on_orthogonal_space(grad, weights, grad, nattrs);
	}

	/* minimization along the gradient (around the center of
	   coordinate system) */
	if(opt.soft_auc || opt.exact || opt.regul)
	{
	    vect_normalize(weights, weights, nattrs);
	    vect_normalize(grad, grad, nattrs);
	}



	if(opt.exact)
	{
	    scale_weights = scale_grad = (FLT_TYPE)1;
	}
	else
	{
	    supports_for_minim_along_grad(&supports_fwfg0, &supports_fwfg1,
					  weights, grad,
					  nattrs, opt.maxdeg, f,
					  &scale_weights,
					  &scale_grad,
					  opt.soft_auc /* use equal
							  scaling in
							  this case,
							  since we are
							  no longer
							  scale-independent */
					  );
	    if(opt.soft_auc)
	    {
		/* we are not scale invariant so we need to regenerate
		   all coefficients. */
		beta_tmp = opt.beta / scale_weights;
		init_cheb_coeff(step_coeff, opt.maxdeg);
		precompute_alphas(&mc, step_coeff, opt.maxdeg);
	    }
	    minim_cache_set_supports(&mc, supports_fwfg0, supports_fwfg1, opt.maxdeg);
	}

	best_alpha = (FLT_TYPE)0;
	best_auc = (FLT_TYPE)-1e30;
	lin_opt_step = 0UL;
	for(alpha = (FLT_TYPE)0; alpha < 2 * PI; alpha += opt.lin_opt_step)
	{
	    FLT_TYPE weights_cur[nattrs];

	    if(opt.exact || opt.regul)
	    {
		for(i = 0; i < nattrs; i++)
		{
		    weights_cur[i] = COS_F(alpha) * weights[i] + SIN_F(alpha) * grad[i];
		}
	    }


	    if(!opt.exact)
	    {
		auc = compute_AUC_along_grad(supports_fwfg0, supports_fwfg1, &mc,
					     alpha, nattrs, opt.maxdeg, lin_opt_step);
	    }
	    else
	    {
		if(opt.soft_auc)
		{
		    beta_tmp = opt.beta;
		    auc = compute_exact_soft_auc(f, weights_cur, nattrs, sigmoid) * N0 * N1;
		}
		else
		{
		    auc = compute_exact_auc(f, weights_cur, nattrs) * N0 * N1;
		}
	    }

	    if(opt.regul)
	    {
		FLT_TYPE regul = (FLT_TYPE)0;	/* regularization penalty */
		
		regul = 0;
		vect_normalize(weights_cur, weights_cur, nattrs);
		for(i = 0; i < nattrs; i++)
		{
		    weights_cur[i] *= 1;
		}

		if(opt.rumelhart)
		{
		    for(i = 0; i < nattrs; i++)
		    {
			FLT_TYPE tmp = weights_cur[i] * weights_cur[i];
			regul += tmp / ((FLT_TYPE)1 + tmp);
		    }
		}
		else if(opt.laplace)
		{
		    for(i = 0; i < nattrs; i++)
		    {
			FLT_TYPE tmp = ABS_F(weights_cur[i]);
			regul += tmp;
		    }
		}
		auc -= opt.lambda * regul * N0 * N1;
		//!auc = -opt.lambda * regul;
	    }
	    //printf("alpha = "FMT",  approx. auc = "PFMT"\n", alpha, auc / N0 / N1);

	    if(auc > best_auc)
	    {
		best_alpha = alpha;
		best_auc = auc;
	    }
	    lin_opt_step++;
	}
	printf("approx. auc = "PFMT"\n", best_auc / N0 / N1);

	if(!opt.exact)
	{
	    free(supports_fwfg1);
	    free(supports_fwfg0);
	}

	/* update the weights */
	for(i = 0; i < nattrs; i++)
	{
	    weights[i] = COS_F(best_alpha) * weights[i] * scale_weights
		+ SIN_F(best_alpha) * grad[i] * scale_grad;
	}

	/* stopping condition - not the nicest one but should work */
	//if(iter > 0 && best_auc == prev_auc)
	if(best_alpha == (FLT_TYPE)0)
	{
	    break;
	}
	prev_auc = best_auc;
    }

    end_time = clock();
    printf("\nCPU time: %f\n", ((double) (end_time - start_time)) / CLOCKS_PER_SEC);

    printf("approx AUC = "PFMT"\n", best_auc / N0 / N1);
    /* print the exact value */
    {
	FLT_TYPE exact_auc;

	if(opt.soft_auc)
	{
	    beta_tmp = opt.beta;
	    vect_normalize(weights, weights, nattrs);
	    exact_auc = compute_exact_soft_auc(f, weights, nattrs, sigmoid);
	    //exact_auc = compute_exact_auc(f, weights, nattrs);
	}
	else
	{
	    exact_auc = compute_exact_auc(f, weights, nattrs);
	}
	printf("exact AUC = "PFMT"\n", exact_auc);
    }    

    /* bring weights back to the original dataset's scaling */
    normalize_vect(weights, nattrs);
    if(opt.print_weights)
    {
	vect_print(weights, nattrs);
    }

    /* test set */
    free_default_cache();	/* free file data cache */
    reset_attribute_normalization(nattrs);
    if(opt.test_fname != NULL)
    {
	FILE * ftest;
	FLT_TYPE exact_auc_test;

	printf("\n ====   test file   ==== \n\n");
	ftest = fopen(opt.test_fname, "r");
	if(ftest == NULL)
	{
	    printf("Cannot open test file %s\n\n", argv[1]);
	}
	else
	{
	    if(opt.soft_auc)
	    {
		exact_auc_test = compute_exact_soft_auc(ftest, weights, nattrs, sigmoid);
		//exact_auc_test = compute_exact_auc(ftest, weights, nattrs);
	    }
	    else
	    {
		exact_auc_test = compute_exact_auc(ftest, weights, nattrs);
	    }
	    printf("exact AUC = "PFMT"\n", exact_auc_test);
	    fclose(ftest);
	}	
    }

    free_attribute_normalization();
    free_minim_cache(&mc);
    free_arrays(step_coeff, weights, grad);

    fclose(f);
    return 0;
}
