#include "exact_auc.h"

#include <stdlib.h>


static int
compare_flts(const void *a, const void *b)
{
    const FLT_TYPE *da = (const FLT_TYPE *) a;
    const FLT_TYPE *db = (const FLT_TYPE *) b;
    
    return (*da > *db) - (*da < *db);
}



FLT_TYPE
compute_exact_auc(FILE * f, FLT_TYPE * weights, LONG_INT_TYPE nattrs)
{
    FLT_TYPE * X;		/* input records vector */
    int class;

    /* store values of f belonging to both classes */
    FLT_TYPE * class0_values;
    FLT_TYPE * class1_values;
    LONG_INT_TYPE class0_alloc;	/* how many values are allocated */
    LONG_INT_TYPE class1_alloc;
    LONG_INT_TYPE class0_N = 0UL;	/* how many values are actually stored */
    LONG_INT_TYPE class1_N = 0UL;

    FLT_TYPE auc = (FLT_TYPE)0;


    X = malloc(nattrs * sizeof(FLT_TYPE));

    /* initial allocation */
    class0_alloc = 128UL;
    class0_values = malloc(class0_alloc * sizeof(FLT_TYPE));
    class1_alloc = 128UL;
    class1_values = malloc(class1_alloc * sizeof(FLT_TYPE));

    /* read the data */
    rewind(f);
    while((class = read_and_normalize_record(f, X, nattrs)) != -1)
    {
	FLT_TYPE f_w;

	f_w = compute_f(weights, X, nattrs);
	if(class == 0)
	{
	    if(class0_N == class0_alloc)
	    {
		class0_alloc *= 2;
		class0_values = realloc(class0_values,
					class0_alloc * sizeof(FLT_TYPE));
	    }
	    class0_values[class0_N] = f_w;
	    class0_N++;
	}
	else
	{
	    if(class1_N == class1_alloc)
	    {
		class1_alloc *= 2;
		class1_values = realloc(class1_values,
					class1_alloc * sizeof(FLT_TYPE));
	    }
	    class1_values[class1_N] = f_w;
	    class1_N++;
	}
    }

    if(class0_N != 0UL && class1_N != 0UL)
    {
	LONG_INT_TYPE i, j;

	qsort(class0_values, class0_N, sizeof(FLT_TYPE), compare_flts);
	qsort(class1_values, class1_N, sizeof(FLT_TYPE), compare_flts);

	j = 0UL;
	for(i = 0UL; i < class0_N; i++)
	{
	    while(j < class1_N && class1_values[j] < class0_values[i])
	    {
		j ++;
	    }
	    auc += class1_N - j;
	}
	auc /= class0_N * class1_N;
    }
    else
    {
	auc = (FLT_TYPE)1;
    }

    free(class1_values);
    free(class0_values);
    free(X);
    return auc;
}

FLT_TYPE
compute_exact_soft_auc(FILE * f, FLT_TYPE * weights, LONG_INT_TYPE nattrs,
		       FLT_TYPE (*sigmoid)(FLT_TYPE x))
{
    FLT_TYPE * X;		/* input records vector */
    int class;

    /* store values of f belonging to both classes */
    FLT_TYPE * class0_values;
    FLT_TYPE * class1_values;
    LONG_INT_TYPE class0_alloc;	/* how many values are allocated */
    LONG_INT_TYPE class1_alloc;
    LONG_INT_TYPE class0_N = 0UL;	/* how many values are actually stored */
    LONG_INT_TYPE class1_N = 0UL;

    FLT_TYPE auc = (FLT_TYPE)0;


    X = malloc(nattrs * sizeof(FLT_TYPE));

    /* initial allocation */
    class0_alloc = 128UL;
    class0_values = malloc(class0_alloc * sizeof(FLT_TYPE));
    class1_alloc = 128UL;
    class1_values = malloc(class1_alloc * sizeof(FLT_TYPE));

    /* read the data */
    rewind(f);
    while((class = read_and_normalize_record(f, X, nattrs)) != -1)
    {
	FLT_TYPE f_w;

	f_w = compute_f(weights, X, nattrs);
	if(class == 0)
	{
	    if(class0_N == class0_alloc)
	    {
		class0_alloc *= 2;
		class0_values = realloc(class0_values,
					class0_alloc * sizeof(FLT_TYPE));
	    }
	    class0_values[class0_N] = f_w;
	    class0_N++;
	}
	else
	{
	    if(class1_N == class1_alloc)
	    {
		class1_alloc *= 2;
		class1_values = realloc(class1_values,
					class1_alloc * sizeof(FLT_TYPE));
	    }
	    class1_values[class1_N] = f_w;
	    class1_N++;
	}
    }

    if(class0_N != 0UL && class1_N != 0UL)
    {
	LONG_INT_TYPE i, j;

	for(i = 0UL; i < class0_N; i++)
	{
	    for(j = 0UL; j < class1_N; j++)
	    {
		auc += sigmoid(class1_values[j] - class0_values[i]);
	    }
	}
	auc /= class0_N * class1_N;
    }
    else
    {
	auc = (FLT_TYPE)1;
    }

    free(class1_values);
    free(class0_values);
    free(X);
    return auc;
}

