#include "minim_along_grad.h"

/* compute supports needed to minimize along the gradient.

  Pointers to support arrays are put into supports_fwfg[01]_ret.
  Scaling for fw and fg is put into scale_w and scale_g
  respectively. */
void
supports_for_minim_along_grad(FLT_TYPE ** supports_fwfg0_ret,
			      FLT_TYPE ** supports_fwfg1_ret,
			      const FLT_TYPE * weights,
			      const FLT_TYPE * grad,
			      size_t nattrs, size_t maxdeg,
			      FILE * f,
			      FLT_TYPE * scale_w, FLT_TYPE * scale_g,
			      int equal_scales)
{
    FLT_TYPE * X;
    FLT_TYPE * supports_fwfg0;	/* supports of powers of f_weights(x) * f_grad(x) */
    FLT_TYPE * supports_fwfg1;	/* supports of powers of f_weights(x) * f_grad(x) */
    FLT_TYPE * supports_fwfg;	/* supports of powers of
				   f_weights(x)*f_grad(x) for current
				   class*/

    int class;
    size_t i, j;
    size_t ind;
    size_t array_size;		/* size of memory allocated for upper
				   triangular array of supports. */

    FLT_TYPE fg_pow;
    FLT_TYPE fgw_pow;

    /* which classes have we seen so far */
    int do_rescale;

    /* used for scaling */
    struct minmax mm_weights;
    struct minmax mm_grad;
    struct minmax mm_common;	/* for equal_scales */

    X = malloc(sizeof(FLT_TYPE) * nattrs);

    /* allocate memory for supports */
    array_size = (maxdeg + 1) * (maxdeg + 2) / 2;
    supports_fwfg0 = malloc(sizeof(FLT_TYPE) * array_size);
    supports_fwfg1 = malloc(sizeof(FLT_TYPE) * array_size);
    for(i = 0; i < array_size; i++)
    {
	supports_fwfg0[i] = (FLT_TYPE)0;
	supports_fwfg1[i] = (FLT_TYPE)0;
    }

    /* compute supports needed for gradient */
    rewind(f);
    init_minmax(&mm_weights);
    init_minmax(&mm_grad);
    init_minmax(&mm_common);

    while((class = read_and_normalize_record(f, X, nattrs)) != -1)
    {
	FLT_TYPE fw, fg;
	/* how much to rescale supports during current iteration to
	   guarantee the [-1, 1] range of cos(alpha)f_w +
	   sin(alpha)f_g */
	FLT_TYPE rescale_factor_weights, rescale_factor_grad;

	/* powers of the rescaling factor used to scale current values
	   of supports */
	FLT_TYPE rescale_pow, rescale_pow_2;


	fw = compute_f(weights, X, nattrs);
	fg = compute_f(grad, X, nattrs);


	/* update maximums and minimums needed for update */
	if(equal_scales)
	{
	    /* use only mm_weights for the common scale */
	    /* ugly! break scaling encapsulation */
	    (void)update_minmax(&mm_weights, fw, class);
	    (void)update_minmax(&mm_grad, fg, class);
	    do_rescale = update_minmax(&mm_common, HYPOT_F(mm_weights.max_M, mm_grad.max_M), class);

	    rescale_factor_weights = mm_common.rescale_factor;
	    rescale_factor_grad = mm_common.rescale_factor;
	    fw *= mm_common.scale_factor;
	    fg *= mm_common.scale_factor;
	}
	else
	{
	    do_rescale = update_minmax(&mm_weights, fw * SQRT_2, class); /* a trick to get the max multiplied by sqrt(2) */
	    do_rescale = do_rescale || update_minmax(&mm_grad, fg * SQRT_2, class);
	    rescale_factor_weights = mm_weights.rescale_factor;
	    rescale_factor_grad = mm_grad.rescale_factor;
	    fw *= mm_weights.scale_factor;
	    fg *= mm_grad.scale_factor;
	}


	/* compute supports */

	/* rescale supports for both classes */
	if(do_rescale)
	{
	    ind = 0;
	    rescale_pow = (FLT_TYPE)1;

	    for(i = 0; i <= maxdeg; i++)
	    {
		rescale_pow_2 = rescale_pow;

		for(j = 0; j <= maxdeg - i; j++)
		{
		    supports_fwfg0[ind] *= rescale_pow_2;
		    supports_fwfg1[ind] *= rescale_pow_2;
		    ind++;
		    rescale_pow_2 *= rescale_factor_weights;
		}
		rescale_pow *= rescale_factor_grad;
	    }
	}

	/* update supports for current class */
	if(class == 0)
	{
	    supports_fwfg = supports_fwfg0;
	}
	else
	{
	    supports_fwfg = supports_fwfg1;
	}
	ind = 0;
	fg_pow = (FLT_TYPE)1;
	for(i = 0; i <= maxdeg; i++)
	{
	    fgw_pow = fg_pow;
	
	    for(j = 0; j <= maxdeg - i; j++)
	    {
		supports_fwfg[ind] += fgw_pow;
		ind++;
		fgw_pow *= fw;
	    }
	    fg_pow *= fg;
	}
    }

    free(X);

    if(equal_scales)
    {
	*scale_w = mm_common.scale_factor;
	*scale_g = mm_common.scale_factor;
    }
    else
    {
	*scale_w = mm_weights.scale_factor;
	*scale_g = mm_grad.scale_factor;
    }

    *supports_fwfg0_ret = supports_fwfg0;
    *supports_fwfg1_ret = supports_fwfg1;
}


/* print the supports needed for minimization along gradient */
void
print_supports_fwfg(FLT_TYPE * supports_fwfg,
		    size_t nattrs, size_t maxdeg)
{
    size_t ind = 0;
    size_t i, j;

    for(i = 0; i <= maxdeg; i++)
    {
	for(j = 0; j <= maxdeg - i; j++)
	{
	    printf("fw^%ldfg^%ld --> "FMT"\n", j, i, supports_fwfg[ind]);
	    ind++;
	}
    }
}

/* compute AUC at angle alpha, based on precumputed supports */
FLT_TYPE
compute_AUC_along_grad(FLT_TYPE * supports_fwfg0,
		       FLT_TYPE * supports_fwfg1,
		       const struct minim_cache * mc,
		       FLT_TYPE alpha,
		       size_t nattrs, size_t maxdeg,
		       size_t lin_opt_step)
{
    size_t i, j, k, l;
    size_t idx;

    FLT_TYPE auc;
    FLT_TYPE * supports0, * supports1;
    FLT_TYPE * sincos = mc -> sincos + lin_opt_step * (maxdeg + 2) * (maxdeg + 1) / 2;

    /* compute supports of powers f_(cos(alpha)w + sin(alpha)g)^k */
    supports0 = malloc(sizeof(FLT_TYPE) * (maxdeg + 1));
    supports1 = malloc(sizeof(FLT_TYPE) * (maxdeg + 1));

    for(i = 0; i <= maxdeg; i++)
    {
	supports0[i] = (FLT_TYPE)0;
	supports1[i] = (FLT_TYPE)0;
    }

    idx = 0UL;
    for(i = 0; i <= maxdeg; i++)
    {
        for(j = 0; j <= i; j++)
    	{
            supports0[i] += sincos[idx] * mc -> supports_fwfg0[idx];
            supports1[i] += sincos[idx] * mc -> supports_fwfg1[idx];
	    idx++;
    	}
    }

    /* approximate the AUC */
    auc = (FLT_TYPE)0;
    idx = 0UL;
    for(k = 0; k <= maxdeg; k++)
    {
	for(l = 0; l <= k; l++)
	{
	    auc += mc -> alphas[idx++] * supports0[k-l] * supports1[l];
	}
    }

    free(supports1);
    free(supports0);

    return auc;
}


/* caching */

void
init_minim_cache(struct minim_cache * mc, size_t maxdeg,
		 FLT_TYPE lin_opt_step, FLT_TYPE * step_coeff)
{
    size_t i, j, idx;
    LONG_INT_TYPE nck;

    mc -> maxdeg = maxdeg;
    mc -> step_coeff = step_coeff;

    /* precompute combinations */
    mc -> cmb = malloc((maxdeg + 2) * (maxdeg + 1) / 2 * sizeof(FLT_TYPE));
    idx = 0UL;
    for(i = 0; i <= maxdeg; i++)
    {
	nck = (LONG_INT_TYPE)1;

	for(j = 0; j <= i; j++)
	{
	    mc -> cmb[idx++] = (FLT_TYPE)nck;
	    nck = nck * (i - j) / (j + 1);
	}
    }

    /* precompute alphas */
    mc -> alphas = NULL;
    precompute_alphas(mc, step_coeff, maxdeg);

    /* precompute sincos */
    {
	size_t n_lin_opt_steps = (size_t)floor(2 * PI / lin_opt_step) + 1;
	size_t step = 0UL;
	FLT_TYPE alpha;
	FLT_TYPE sa, ca;
	FLT_TYPE * sa_pow, * ca_pow;
	FLT_TYPE sa_pow_tmp, ca_pow_tmp;

	mc -> sincos = malloc(n_lin_opt_steps * (maxdeg + 2) * (maxdeg + 1) / 2 * sizeof(FLT_TYPE));

	sa_pow = malloc(sizeof(FLT_TYPE) * (maxdeg + 1));
	ca_pow = malloc(sizeof(FLT_TYPE) * (maxdeg + 1));


	for(alpha = (FLT_TYPE)0; alpha < 2 * PI; alpha += lin_opt_step)
	{
	    sa = SIN_F(alpha);
	    ca = COS_F(alpha);

	    sa_pow_tmp = (FLT_TYPE)1;
	    ca_pow_tmp = (FLT_TYPE)1;
	    for(i = 0; i <= maxdeg; i++)
	    {
		sa_pow[i] = sa_pow_tmp;
		sa_pow_tmp *= sa;
		ca_pow[i] = ca_pow_tmp;
		ca_pow_tmp *= ca;
	    }

	    idx = 0UL;
	    for(i = 0; i <= maxdeg; i++)
	    {
		for(j = 0; j <= i; j++)
		{
		    mc -> sincos[step * (maxdeg + 2) * (maxdeg + 1) / 2 + idx] = ca_pow[j] * sa_pow[i-j];
		    idx++;
		}
	    }
	    step++;
	}

	free(ca_pow);
	free(sa_pow);
    }

    mc -> supports_fwfg0 = NULL;
    mc -> supports_fwfg1 = NULL;
}

void
precompute_alphas(struct minim_cache * mc, FLT_TYPE * step_coeff, size_t maxdeg)
{
    size_t k, l;
    size_t idx = 0UL;

    /* free old alphas if necessary */
    if(mc -> alphas != NULL)
    {
	free(mc -> alphas);
    }
    mc -> alphas = malloc((maxdeg + 2) * (maxdeg + 1) / 2 * sizeof(FLT_TYPE));

    for(k = 0; k <= maxdeg; k++)
    {
	FLT_TYPE ck = step_coeff[k];

	for(l = 0; l <= k; l++)
	{
	    FLT_TYPE alpha_kl = ck * mc -> cmb[idx];

	    if((k - l) % 2 != 0)
	    {
		alpha_kl = -alpha_kl;
	    }
	    mc -> alphas[idx++] = alpha_kl;
	}
    }
}

void minim_cache_set_supports(struct minim_cache * mc,
			      FLT_TYPE * supports_fwfg0,
			      FLT_TYPE * supports_fwfg1,
			      size_t maxdeg)
{
    size_t i, j, idx;
    
    /* free old supports */
    if(mc -> supports_fwfg0 != NULL)
    {
	free(mc -> supports_fwfg0);
    }
    if(mc -> supports_fwfg1 != NULL)
    {
	free(mc -> supports_fwfg1);
    }

    /* create new cached supports */
    mc -> supports_fwfg0 = malloc((maxdeg + 2) * (maxdeg + 1) / 2 * sizeof(FLT_TYPE));
    mc -> supports_fwfg1 = malloc((maxdeg + 2) * (maxdeg + 1) / 2 * sizeof(FLT_TYPE));

    idx = 0UL;
    for(i = 0; i <= maxdeg; i++)
    {
        for(j = 0; j <= i; j++)
	{
	    FLT_TYPE cmb = mc -> cmb[idx];
	    /* index of [i - j][j] for triangular matrix */
	    size_t ind = ((i-j) * (2 * (maxdeg + 1) - (i-j) + 1)) / 2 + j;

            mc -> supports_fwfg0[idx] = cmb * supports_fwfg0[ind];
            mc -> supports_fwfg1[idx] = cmb * supports_fwfg1[ind];
	    idx++;
	}
    }

}

void
free_minim_cache(struct minim_cache * mc)
{
    free(mc -> sincos);
    free(mc -> alphas);
    free(mc -> cmb);
    if(mc -> supports_fwfg0 != NULL)
    {
	free(mc -> supports_fwfg0);
    }
    if(mc -> supports_fwfg1 != NULL)
    {
	free(mc -> supports_fwfg1);
    }
}
